/*******************************************************************************
  Main Source File

  Company:
    Microchip Technology Inc.

  File Name:
    main.c

  Summary:
    This file contains the "main" function for a project.

  Description:
    This file contains the "main" function for a project.  The
    "main" function calls the "SYS_Initialize" function to initialize the state
    machines of all modules in the system
 *******************************************************************************/

// *****************************************************************************
// *****************************************************************************
// Section: Included Files
// *****************************************************************************
// *****************************************************************************

#include <stddef.h>                     // Defines NULL
#include <stdbool.h>                    // Defines true
#include <stdlib.h>                     // Defines EXIT_FAILURE
#include "definitions.h"                // SYS function prototypes

#include "bmi160.h"
#include "utils.h"
#include "edge-impulse-sdk/classifier/ei_run_classifier.h"

// *****************************************************************************
// *****************************************************************************
// Section: Build Configuration Macros
// *****************************************************************************
// *****************************************************************************

/*  DATA_STREAMER_BUILD = 1 for streaming IMU data to MPLAB Data Visualizer
                        = 0 for impulse deployment (embedded inference)
 */
#define DATA_STREAMER_BUILD 0

#if DATA_STREAMER_BUILD == 0
#define IMPULSE_DEPLOYMENT_BUILD 1
#endif

// *****************************************************************************
// *****************************************************************************
// Section: Local Variables
// *****************************************************************************
// *****************************************************************************

struct bmi160_dev sensor;
struct bmi160_sensor_data accel;
struct bmi160_sensor_data gyro;

volatile int8_t sensorstatus;

#if IMPULSE_DEPLOYMENT_BUILD

struct sensor_buffer_t {
                float       data[EI_CLASSIFIER_DSP_INPUT_FRAME_SIZE];
    volatile    uint32_t    write_index;
    volatile    bool        buffer_full;
} sensorbuffer;

//struct sensor_buffer_t sensorbuffer;

#endif //IMPULSE_DEPLOYMENT_BUILD

// *****************************************************************************
// *****************************************************************************
// Section: Local Functions
// *****************************************************************************
// *****************************************************************************

void EIC_Callback(unsigned int context);

#if IMPULSE_DEPLOYMENT_BUILD

int get_feature_data(size_t offset, size_t length, float *out_ptr) {
    memcpy(out_ptr, sensorbuffer.data + offset, length * sizeof(float));
    return 0;
}

#endif //IMPULSE_DEPLOYMENT_BUILD

// *****************************************************************************
// *****************************************************************************
// Section: Main Entry Point
// *****************************************************************************
// *****************************************************************************

int main ( void )
{
    /* Initialize all modules */
    SYS_Initialize ( NULL );
    SYSTICK_TimerStart();
    SYSTICK_DelayMs(500u);

#if IMPULSE_DEPLOYMENT_BUILD    
    /* Edge Impulse - Initialize Signal */
    signal_t signal;
    signal.total_length = sizeof(sensorbuffer.data) / sizeof(sensorbuffer.data[0]);
    signal.get_data = &get_feature_data;
    
    bool uncertain = true;
    
    memset(&sensorbuffer, 0, sizeof(struct sensor_buffer_t));
#endif
    
    /* Initialize BMI160 */
    sensor.id = BMI160_I2C_ADDR;
    sensor.interface = BMI160_I2C_INTF;
    sensor.read = user_i2c_read;
    sensor.write = user_i2c_write;
    sensor.delay_ms = SYSTICK_DelayMs;
    
    int8_t rslt = BMI160_OK;
    rslt = bmi160_init(&sensor);
    printf("init result = %d\n",rslt);
    
    /* Select the Output data rate, range of accelerometer sensor */
    sensor.accel_cfg.odr = BMI160_ACCEL_ODR_100HZ;
    sensor.accel_cfg.range = BMI160_ACCEL_RANGE_2G;
    sensor.accel_cfg.bw = BMI160_ACCEL_BW_NORMAL_AVG4;

    /* Select the power mode of accelerometer sensor */
    sensor.accel_cfg.power = BMI160_ACCEL_NORMAL_MODE;

    /* Select the Output data rate, range of Gyroscope sensor */
    sensor.gyro_cfg.odr = BMI160_GYRO_ODR_100HZ;
    sensor.gyro_cfg.range = BMI160_GYRO_RANGE_2000_DPS;
    sensor.gyro_cfg.bw = BMI160_GYRO_BW_NORMAL_MODE;

    /* Select the power mode of Gyroscope sensor */
    sensor.gyro_cfg.power = BMI160_GYRO_NORMAL_MODE; 

    /* Set the sensor configuration */
    rslt = BMI160_OK;
    rslt = bmi160_set_sens_conf(&sensor);
    printf("configuration result = %d\n",rslt);
    
    /* Initialize status and our data buffer */
    sensorstatus = BMI160_OK;
    
    /* Interrupt Configuration */
    struct bmi160_int_settg int_config;

    /* Select the Interrupt channel/pin */
    int_config.int_channel = BMI160_INT_CHANNEL_1;// Interrupt channel/pin 1

    /* Select the Interrupt type */
    int_config.int_type = BMI160_ACC_GYRO_DATA_RDY_INT;// Choosing data ready interrupt
    
    /* Select the interrupt channel/pin settings */
    int_config.int_pin_settg.output_en = BMI160_ENABLE;// Enabling interrupt pins to act as output pin
    int_config.int_pin_settg.output_mode = BMI160_DISABLE;// Choosing push-pull mode for interrupt pin
    int_config.int_pin_settg.output_type = BMI160_DISABLE;// Choosing active low output
    int_config.int_pin_settg.edge_ctrl = BMI160_ENABLE;// Choosing edge triggered output
    int_config.int_pin_settg.input_en = BMI160_DISABLE;// Disabling interrupt pin to act as input
    int_config.int_pin_settg.latch_dur = BMI160_LATCH_DUR_NONE;// non-latched output
            
    /* Set the data ready interrupt */
    rslt = BMI160_OK;
    rslt = bmi160_set_int_config(&int_config, &sensor);
    printf("interrupt configuration result = %d\n",rslt);
    
    SERCOM_I2C_ERROR i2cstatus;
    USART_ERROR uartstatus;
    
    RTC_Timer32Start();
    Activate_EIC();
    
    while ( true )
    {
        if (sensorstatus != BMI160_OK) {
            printf("Got a bad sensor status: %d\n", sensorstatus);
            break;
        }
        else if ((i2cstatus = SERCOM1_I2C_ErrorGet()) != SERCOM_I2C_ERROR_NONE) {
            printf("i2c Error: %d\n", i2cstatus);
            break;
        }
        else if ((uartstatus = SERCOM5_USART_ErrorGet()) & ~(USART_ERROR_NONE | USART_ERROR_OVERRUN)) {
            printf("uart Error: %d\n", uartstatus);
            break;
        }
        else {
            
#if IMPULSE_DEPLOYMENT_BUILD
            if (sensorbuffer.buffer_full) {

                printf("\n\nRunning Impulse...");
                RTC_Timer32CounterSet(0u);
                ei_impulse_result_t result = { 0 };
                EI_IMPULSE_ERROR res = run_classifier(&signal, &result, false);
                if (res!=0) {
                    printf("\n\n\n\trun_classifier() FAILED!\n\n\terror code: %d\n\n\n",res);
                    return 1;
                }
                
                printf("\nTimings (DSP: %d ms, Classifier: %d ms, Anomaly: %d ms)",
                        result.timing.dsp, result.timing.classification, result.timing.anomaly);

#if EI_CLASSIFIER_HAS_ANOMALY == 1
                if (result.anomaly > 0.7) {
                    printf("\n  -> *ANOMALY*");
                    //printf(" (%.2f)", result.anomaly);
                } else {
                    for (size_t ix = 0; ix < EI_CLASSIFIER_LABEL_COUNT; ix++) {
                        if (result.classification[ix].value > 0.85) {
                            printf("\n  -> %s", result.classification[ix].label);
//                            printf(" (%.2f)", result.classification[ix].value);
                            uncertain = false;
                        }
                    }
                    if (uncertain) {
                        printf("\n  -> UNCERTAIN");
                    }
                }
#else
                for (size_t ix = 0; ix < EI_CLASSIFIER_LABEL_COUNT; ix++) {
                    if (result.classification[ix].value > 0.85) {
                        printf("\n    -> %s", result.classification[ix].label);
//                        printf(" (%.2f)", result.classification[ix].value);
                    }
                }                
#endif                
                uncertain = true;
                sensorbuffer.buffer_full = false;    
            }
#endif //IMPULSE_DEPLOYMENT_BUILD
        }
    }

    /* Execution should not come here during normal operation */

    return ( EXIT_FAILURE );
}


void EIC_Callback(unsigned int context) {
    /* Check if any errors we've flagged have been acknowledged */
    if (sensorstatus != BMI160_OK) {
        return;
    }
    
#if DATA_STREAMER_BUILD
    
    /* To read both Accel and Gyro data */
    sensorstatus = bmi160_get_sensor_data((BMI160_ACCEL_SEL | BMI160_GYRO_SEL), &accel, &gyro, &sensor);
    if (sensorstatus != BMI160_OK)
        return;
    
    // Transmit Start Token
    SERCOM5_USART_WriteByte(0x55);
    
    // Transmit Data
    SERCOM5_USART_WriteByte(accel.x);
    SERCOM5_USART_WriteByte(accel.x >> 8);
    SERCOM5_USART_WriteByte(accel.y);
    SERCOM5_USART_WriteByte(accel.y >> 8);
    SERCOM5_USART_WriteByte(accel.z);
    SERCOM5_USART_WriteByte(accel.z >> 8);
    SERCOM5_USART_WriteByte(gyro.x);
    SERCOM5_USART_WriteByte(gyro.x >> 8);
    SERCOM5_USART_WriteByte(gyro.y);
    SERCOM5_USART_WriteByte(gyro.y >> 8);
    SERCOM5_USART_WriteByte(gyro.z);
    SERCOM5_USART_WriteByte(gyro.z >> 8);
    
    // Transmit End Token
    SERCOM5_USART_WriteByte(0xAA);
    
#elif IMPULSE_DEPLOYMENT_BUILD

    /* Check for overrun first */
    if (sensorbuffer.write_index >= EI_CLASSIFIER_DSP_INPUT_FRAME_SIZE) {
        return;
    }
    
    /* To read both Accel and Gyro data */
    sensorstatus = bmi160_get_sensor_data((BMI160_ACCEL_SEL | BMI160_GYRO_SEL), &accel, &gyro, &sensor);
    if (sensorstatus != BMI160_OK)
        return;
    
    sensorbuffer.data[sensorbuffer.write_index++] = (float)accel.x;
    sensorbuffer.data[sensorbuffer.write_index++] = (float)accel.y;
    sensorbuffer.data[sensorbuffer.write_index++] = (float)accel.z;
    sensorbuffer.data[sensorbuffer.write_index++] = (float)gyro.x;
    sensorbuffer.data[sensorbuffer.write_index++] = (float)gyro.y;
    sensorbuffer.data[sensorbuffer.write_index++] = (float)gyro.z;
    
    if (sensorbuffer.write_index == EI_CLASSIFIER_DSP_INPUT_FRAME_SIZE) {
        sensorbuffer.buffer_full = true;
        sensorbuffer.write_index = 0;
    }
#endif    
}

/*******************************************************************************
 End of File
*/

